class TruffleCompiler:
    """Truffle compiler interface exposes compiling functionality from truffle"""
    def run_compile_command(self, working_directory: str) -> None:
        """ Runs truffle's test command

        :param working_directory: The truffle project directory to compile
        """
        raise NotImplementedError



from eth_vertigo.test_runner.truffle.truffle_tester import TruffleTester
from eth_vertigo.mutation.truffle.truffle_compiler import TruffleCompiler

from eth_vertigo.test_runner.exceptions import TestRunException, TimedOut
from eth_vertigo.test_runner import TestResult
from json import loads, JSONDecodeError
from subprocess import Popen, TimeoutExpired
from tempfile import TemporaryFile

from typing import Dict, Union


class Truffle(TruffleTester, TruffleCompiler):
    """ Truffle interface object, deals with the ugly commandline details"""

    def __init__(self, truffle_location: str = "truffle") -> None:
        """ Initializes a new truffle object

        :param truffle_location: Location where the truffle cli can be found
        """
        self.truffle_location = truffle_location

    def run_compile_command(self, working_directory: str) -> None:
        command = [
            self.truffle_location, 'compile'
        ]

        with TemporaryFile() as stdin, TemporaryFile() as stdout:
            stdin.seek(0)
            proc = Popen(command, stdin=stdin, stdout=stdout, cwd=working_directory)
            proc.wait()
            stdout.seek(0)
            output = stdout.read()

        split = output.decode('utf-8').split("\n")

        errors = []
        for line in split:
            if line.startswith("Error"):
                errors.append(line)

        if errors:
            raise Exception("Encountered compilation error: \n" + "\n".join(errors))

    def run_test_command(self, working_directory: str, file_name: str = None, network_name: str = None, timeout=None) -> Union[Dict[str, TestResult], None]:
        command = [
            self.truffle_location, 'test'
        ]

        if file_name:
            command.append(file_name)
        if network_name:
            command += ['--network', network_name]

        with TemporaryFile() as stdin, TemporaryFile() as stdout:
            stdin.seek(0)
            proc = Popen(command, stdin=stdin, stdout=stdout, cwd=working_directory)
            try:
                proc.wait(timeout=timeout)
            except TimeoutExpired:
                proc.kill()
                raise TimedOut

            stdout.seek(0)
            output = stdout.read()

        split = output.decode('utf-8').split("\n")
        errors = []
        test_result = []
        preamble = True
        for line in split:
            if line.startswith("Error"):
                errors.append(line)
            if line.startswith("{"):
                preamble = False
            if preamble:
                continue
            test_result.append(line)

        test_result = "\n".join(test_result)

        if errors:
            raise TestRunException("\n".join(errors))
        try:
            return self._normalize_mocha(loads(test_result))
        except JSONDecodeError:
            raise TestRunException("Encountered error during test output analysis")

    @staticmethod
    def _normalize_mocha(mocha_json: dict) -> Dict[str, TestResult]:
        tests = {}
        for failure in mocha_json["failures"]:
            tests[failure["fullTitle"]] = TestResult(failure["title"], failure["fullTitle"], failure["duration"], False)
        for success in mocha_json["passes"]:
            tests[success["fullTitle"]] = TestResult(success["title"], success["fullTitle"], success["duration"], True)
        return tests



from eth_vertigo.source.truffle.solidity_file import SolidityFile
from eth_vertigo.source.solidity.solidity_mutator import SolidityMutator
from eth_vertigo.test_runner.truffle import TruffleRunnerFactory
from eth_vertigo.mutation import Mutation, MutationResult
from eth_vertigo.mutation.campaign import Campaign
from eth_vertigo.mutation.truffle.truffle_compiler import TruffleCompiler
from eth_vertigo.test_runner.exceptions import TestRunException, TimedOut

from typing import List, Callable
from time import time
from pathlib import Path
import logging
from queue import Queue


class TruffleCampaign(Campaign):
    """
    A TruffleCampaign class

    Implements specific campaign logic for the truffle framework
    """
    def __init__(self, project_directory: Path, truffle_compiler: TruffleCompiler, truffle_runner_factory: TruffleRunnerFactory, networks: List[str] = (), filters=None):
        super().__init__(filters=filters)
        self.project_directory = project_directory
        self.source_directory = project_directory / "build" / "contracts"
        self.truffle_compiler = truffle_compiler

        self.sources = list(self._get_sources())
        self.base_run_time = None
        self.networks = networks
        self.networks_queue = Queue(maxsize=len(networks))

        self.truffle_runner_factory = truffle_runner_factory

        for network in networks:
            self.networks_queue.put(network)

    def _get_sources(self, dir=None):
        """ Implements basic source file discovery """
        if not (self.project_directory / "build").exists():
            self.truffle_compiler.run_compile_command(str(self.project_directory))

        dir = dir or self.source_directory
        for source_file in dir.iterdir():
            if source_file.name == "Migrations.json":
                continue
            if not source_file.name.endswith(".json"):
                continue
            yield SolidityFile(source_file)

    def valid(self):
        """ Checks whether the current project is valid """
        tr = self.truffle_runner_factory.create(str(self.project_directory))

        begin = time()

        network = None
        if self.networks:
            network = self.networks_queue.get()
        try:
            test_result = tr.run_tests(network=network)
        finally:
            if self.networks:
                self.networks_queue.put(network)

        self.base_run_time = time() - begin
        if test_result is None:
            return False

        return True

    def setup(self):
        mutator = SolidityMutator()
        for source in self.sources:
            self.mutations += mutator.mutate(source, self.project_directory)
        for f in self.filters:
            self.mutations = f.apply(self.mutations)
        self.is_set_up = True

    def test_mutation(self, mutation: Mutation, done_callback: Callable):
        """ Run the test suite using a mutation and check for murders """
        tr = self.truffle_runner_factory.create(str(self.project_directory))
        mutation.result = MutationResult.LIVED
        network = None
        if self.networks:
            network = self.networks_queue.get()
        try:
            test_result = tr.run_tests(mutation=mutation, timeout=int(self.base_run_time) * 2, network=network)
            if any(map(lambda t: not t.success, test_result.values())):
                mutation.result = MutationResult.KILLED
        except TimedOut:
            mutation.result = MutationResult.TIMEDOUT
        except TestRunException as e:
            logging.warning(str(e))
            mutation.result = MutationResult.ERROR
        finally:
            if self.networks:
                self.networks_queue.put(network)
            done_callback()
            return




from eth_vertigo.test_runner.test_result import TestResult
from typing import Dict, Union


class TruffleTester:
    """Truffle tester interface exposes testing functionality from truffle"""
    def run_test_command(self, working_directory: str, file_name: str = None, network_name: str = None, timeout=None) \
            -> Union[None, Dict[str, TestResult]]:
        """ Runs truffle's test command

        :param working_directory: The directory which will be tested
        :param file_name: Filename of the test that should be executed
        :param network_name: Name of the network to execute the tests on
        :param timeout: The maximum duration that the tests are allowed to take
        :return: The test results
        """
        raise NotImplementedError
